// Slickworm
// Copyright (c) 2002-2003, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "graphicssystemfactory.h"
#include "mouse.h"
#include "mtxlib.h"
#include "landscape_collider.h"
#include "keyboard.h"
#include "camera.h"
#include "frustum.h"
#include "ode_thread.h"
#include "ode_helper.h"
#include "filenames.h"
#include "application.h"
#include "landscape_deformation.h"
#include "get_distance.h"
#include "buggy.h"
#include "explosion.h"
#include "first_person.h"
#include "vapour_trail.h"
#include "flare.h"
#include "gettime.h"
#include "sky.h"
#include "sea.h"
#include "audiosystemopenal.h"
#include "hud_renderer.h"
#include "hud.h"
#include "font.h"
#include <string>
#include <sstream>
#include <iostream>
#include <iomanip>
#include "drawstuff.h"
#include "object_renderer.h"

#if defined(WIN32)
#include "SDL_version.h"
#include "SDL_syswm.h"
#endif // defined(WIN32)

Camera cam1(int( 500 * 1.5));
Camera cam2(int(1000 * 1.5));

#if defined(USE_SOUND)
ALuint explosion_sample;
ALuint active_sources[100];
#endif // defined(USE_SOUND)

extern int g_stencil;

// Constructor
Application::Application()
{
    try
    {
        IGS = GraphicsSystemFactory::CreateGraphicsSystem(GraphicsSystemFactory::GRAPHICS_OPENGL);
        hud_misc = IGS->CreateTexture(data_dir() + slash() + "hud_misc.png", false, true);

        this->ShowLoadingScreen(5);

        sky = new Sky(data_dir() + slash() + "sky_top.png",
            data_dir() + slash() + "sky_west.png",
            data_dir() + slash() + "sky_north.png",
            data_dir() + slash() + "sky_east.png",
            data_dir() + slash() + "sky_south.png", IGS);

        this->ShowLoadingScreen(10);

        sea = new Sea(data_dir() + slash() + "sea.png", IGS);

        this->ShowLoadingScreen(15);

        land = new Landscape(data_dir() + slash() + "level001_height.raw",
            data_dir() + slash() + "level001_terrain_image.jpg",
            data_dir() + slash() + "detail.png",
            data_dir() + slash() + "level001_outer.png", IGS);

        this->ShowLoadingScreen(40);

        // Load a fire texture with an alpha channel.
        fire0_texture = IGS->CreateTexture(data_dir() + slash() + "fire0.png", true, true);
        fire1_texture = IGS->CreateTexture(data_dir() + slash() + "fire1.png", true, true);
        fire2_texture = IGS->CreateTexture(data_dir() + slash() + "fire2.png", true, true);
        smoke0_texture= IGS->CreateTexture(data_dir() + slash() + "smoke0.png", true, true);
        smoke1_texture= IGS->CreateTexture(data_dir() + slash() + "smoke1.png", true, true);

        this->ShowLoadingScreen(50);

        font_texture = IGS->CreateTexture(data_dir() + slash() + "font.png", false, true);

        this->ShowLoadingScreen(55);

        dWorldID world = OdeThread::Instance().getWorld();
        dSpaceID space = OdeThread::Instance().getSpace();

        buggy = new Buggy(world, space, IGS, data_dir() + slash() + "buggy" + slash(),
            "buggy_chassis.txt", "wheel4.txt", vector3(172, 24, 396));

        this->ShowLoadingScreen(75);

        fp = new FirstPerson(world, space, vector3(145, 24, 386), vector3(1, 0, 0),
            data_dir() + slash(), "shotgun.txt", IGS);

        this->ShowLoadingScreen(85);

        // Create the geom for the terrain
        landscape_geom = dCreateLandscape(space, land->GetHeightfieldF(), 513, 513);
        dGeomSetBody(landscape_geom, 0);    // Terrain is immovable object. (See ODE FAQ 11.4)

        this->ShowLoadingScreen(90);

        // Create the bodies for the spheres
        #define DENSITY (1.0)   // Density of all objects
        dMass m;

        // Create the sphere bodies
        for (int s = 0; s < 4; s++)
        {
            body[s] = dBodyCreate(world);

            dMassSetSphere(&m, DENSITY, 3.0);
            dMassAdjust (&m, 1.0);
            dBodySetMass(body[s], &m);
            geom[s] = dCreateSphere(space, 3.0);
            dGeomSetBody(geom[s], body[s]);

            dBodySetPosition(body[s], 215.2 + s * 8.0, 30.0, 348.2);
        }




        OdeThread::Instance().start_thread();

        for (int r = 0; r < 30; r++)
        {
            rockets[r].active = false;

            rockets[r].ex = NULL;
            rockets[r].vt = NULL;
            rockets[r].fl = NULL;
        }

        this->ShowLoadingScreen(95);


        beastie = ModelFactory::CreateModel(ModelFactory::MILKSHAPE_ASCII);
        beastie->Load(data_dir() + slash() + "enemies" + slash(), "insect.txt", IGS);
        beastie_pose = beastie->CreateAnimationInstance();

        thing = ModelFactory::CreateModel(ModelFactory::MILKSHAPE_ASCII);
        thing->Load(data_dir() + slash() + "rocks" + slash(), "rock_medium.txt", IGS);

        this->ShowLoadingScreen(100);


#if defined(USE_SOUND)
        al = new OpenAL();

        if (al->Init() == 1)
        {
            audio_available = false;
            std::cout << "OpenAL Initialisation failed" << std::endl;
        }
        else
        {
            audio_available = true;
        }

        if (audio_available)
        {
            alDistanceModel(AL_INVERSE_DISTANCE);

            explosion_sample = al->CreateBuffer();

            if (al->LoadBuffer(explosion_sample,
                data_dir() + slash() + "sound" + slash() + "explosion.ogg") == 1)
            {
                std::cout << "Failed to load ogg file" << std::endl;
                exit(1);
            }

            for (int i = 0; i < 100; i++)
            {
                active_sources[i] = 0;
            }
        }
#endif // defined(USE_SOUND)




    }
    catch (std::string error_message)
    {
#if defined(WIN32)
        SDL_SysWMinfo wm_info;
        wm_info.version = SDL_version();
        if (SDL_GetWMInfo(&wm_info) == 0)
        {
            MessageBox(NULL, error_message.c_str(), "Error message",MB_OK);
        }
        else
        {
            MessageBox(wm_info.window, error_message.c_str(), "Error message",MB_OK);
        }
#endif // defined(WIN32)
        std::cout << "ERROR: " << error_message << std::endl;
        exit(1);
    }
}

// Destructor
Application::~Application()
{
#if defined(USE_SOUND)
    if (audio_available)
    {
        al->DeleteBuffer(explosion_sample);
        
        al->Shutdown();
    }
    delete al;
#endif // defined(USE_SOUND)

    beastie->DeleteAnimationInstance(beastie_pose);
    delete beastie;

    for (int r = 0; r < 30; r++)
    {
        if (rockets[r].ex != NULL) delete rockets[r].ex;
        if (rockets[r].vt != NULL) delete rockets[r].vt;
        if (rockets[r].fl != NULL) delete rockets[r].fl;
    }

    OdeThread::Instance().stop_thread();
    
    delete buggy;

    if (fp != NULL) delete fp;
    
    // Destroy ODE stuff
    dGeomDestroy(landscape_geom);

    for (int s = 0; s < 4; s++)
    {
        dGeomDestroy(geom[s]);
        dBodyDestroy(body[s]);
    }
    delete thing;

    delete land;
    delete sky;
    delete sea;

    IGS->DeleteTexture(fire0_texture);
    IGS->DeleteTexture(fire1_texture);
    IGS->DeleteTexture(fire2_texture);
    IGS->DeleteTexture(smoke0_texture);
    IGS->DeleteTexture(smoke1_texture);
    delete IGS;
}

void Application::ShowLoadingScreen(int percent)
{
    IGS->Clear(COLOUR(0.0f, 0.0f, 0.0f, 1.0f));

    IGS->Begin();

    HUD_Renderer::Instance().Begin(IGS);

    IGS->SetTexture(0, hud_misc);

    draw_health_bar(50, percent / 2, BAR_BLUE, 140.0, 280.0);

    HUD_Renderer::Instance().Draw2D(140.0f, 265.0f, 140.0 + 62.0, 265.0 + 16.0,
        64.0 / 256.0, 32.0 / 256.0, (64.0 + 62.0) / 256.0, 48.0 / 256.0, 
        COLOUR(1.0f, 1.0f, 1.0f, 1.0f));

    HUD_Renderer::Instance().End();

    IGS->End();

    IGS->Display();
}

// Main game Update function. Call this once per frame.
void Application::Update()
{

#if !defined(PHYSICS_IN_SEPARATE_THREAD)
    OdeThread::Instance().Update();
#endif // !defined(PHYSICS_IN_SEPARATE_THREAD)

    // FPS counter
    static long framecount = 0;
    framecount++;
    static long start_time = GetTime();

    long elapsed_time = GetTime() - start_time;

    std::stringstream frame_rate_sstr;
    float frame_rate = 0;
    if (framecount > 0)
    {
        frame_rate = 1000.0f / float((float)elapsed_time / (float)framecount);

        frame_rate_sstr.setf(std::ios::fixed,std::ios::floatfield);
        frame_rate_sstr << std::setprecision(1) << frame_rate << " FPS";
    }

    if (framecount % 10 == 0)
    {
        framecount = 0;
        start_time = GetTime();
    }




    try
    {

        static int screenshot_number = 0;
        if (Keyboard::Instance().KeyPressed(KEY_O))
        {
            screenshot_number++;
            std::stringstream screenshot_name;
            screenshot_name << "slickworm" << screenshot_number << ".ppm";

            IGS->ScreenShot(screenshot_name.str());
        }

        if (Keyboard::Instance().KeyPressed(KEY_X))
        {
            IGS->EnableWireframe(true);
        }
        if (Keyboard::Instance().KeyPressed(KEY_Z))
        {
            IGS->EnableWireframe(false);
        }

        static bool control_vehicle = false;
        static bool return_key_up = true;
        if (!Keyboard::Instance().KeyPressed(KEY_RETURN)) return_key_up = true;

        if (Keyboard::Instance().KeyPressed(KEY_RETURN) && control_vehicle && return_key_up)
        {
            dWorldID world = OdeThread::Instance().getWorld();
            dSpaceID space = OdeThread::Instance().getSpace();

            vector3 spawn_point = buggy->GetPointInWorldSpace(vector3(-8.0, 1.0, -5.0));
            float height = land->GetHeight(spawn_point.x, spawn_point.z);
            if (spawn_point.y < height) spawn_point.y = (height + 4.0);

            fp = new FirstPerson(world, space, spawn_point,
                buggy->GetDirection(), data_dir() + slash(), "shotgun.txt", IGS);

            control_vehicle = false;
            return_key_up = false;
        }
        else if (Keyboard::Instance().KeyPressed(KEY_RETURN) && return_key_up)
        {
            if (get_distance(buggy->GetPosition(), fp->GetPosition()) < 10.0f)
            {
                // Some camera trickery to smoothly interpolate between being on foot and
                // then overlooking the car
                vector3 camera, lookat, up;
                fp->GetCameraPosition(camera, lookat, up);
                cam1.clear();
                cam2.clear();
                cam2.setCurrentPosition(camera);
                cam1.setCurrentPosition(lookat);

                delete fp;
                fp = NULL;

                control_vehicle = true;
                return_key_up = false;
            }
        }
        
        dWorldID world = OdeThread::Instance().getWorld();
        dSpaceID space = OdeThread::Instance().getSpace();

        if (Keyboard::Instance().KeyPressed(KEY_R) || Keyboard::Instance().KeyPressed(KEY_K))
        {
            delete buggy;
            buggy = new Buggy(world, space, IGS, data_dir() + slash() + "buggy" +
                slash(), "buggy_chassis.txt", "wheel4.txt", vector3(172, 24, 396));
        }

        if (fp != NULL && Mouse::Instance().isLeftButtonDown())
        {
            vector3 weapon_position, weapon_direction;
            fp->GetWeaponLocationAndDirection(weapon_position, weapon_direction);

            int r;

            bool too_soon = false;

            for (r = 0; r < 30; r++)
            {
                if (rockets[r].active == true)
                {
                    if (GetTime() - rockets[r].time_fired < 600)
                    {
                        too_soon = true;
                    }
                }
            }


            for (r = 0; r < 30; r++)
            {
                if (too_soon == false && rockets[r].active == false)
                {
                    rockets[r].active = true;
                
                    rockets[r].vt = new VapourTrail(smoke1_texture, weapon_position, weapon_direction*45.0);
                    rockets[r].fl = new Flare(fire2_texture, weapon_position, weapon_direction*45.0);
                    
                    rockets[r].initial_position = weapon_position;
                    rockets[r].velocity = weapon_direction*45.0;

                    rockets[r].time_fired = GetTime();

                    break;
                }
            }
        }



        if (control_vehicle == true)
        {
            int buggy_direction_flags = 0;

            if (Keyboard::Instance().KeyPressed(KEY_UP) || Keyboard::Instance().KeyPressed(KEY_W))
            {
                buggy_direction_flags |= BUGGY_FORWARD;
            }
            if (Keyboard::Instance().KeyPressed(KEY_DOWN) || Keyboard::Instance().KeyPressed(KEY_S))
            {
                buggy_direction_flags |= BUGGY_REVERSE;
            }
            if (Keyboard::Instance().KeyPressed(KEY_LEFT) || Keyboard::Instance().KeyPressed(KEY_A))
            {
                buggy_direction_flags |= BUGGY_TURN_LEFT;
            }
            if (Keyboard::Instance().KeyPressed(KEY_RIGHT) || Keyboard::Instance().KeyPressed(KEY_D))
            {
                buggy_direction_flags |= BUGGY_TURN_RIGHT;
            }

            buggy->Update(buggy_direction_flags);
        }
        else
        {
            int buggy_direction_flags = 0;
            buggy->Update(buggy_direction_flags);

            int fp_direction_flags = 0;

            if (Keyboard::Instance().KeyPressed(KEY_UP) || Keyboard::Instance().KeyPressed(KEY_W))
            {
                fp_direction_flags |= FP_RUN_FORWARDS;
            }
            if (Keyboard::Instance().KeyPressed(KEY_DOWN) || Keyboard::Instance().KeyPressed(KEY_S))
            {
                fp_direction_flags |= FP_RUN_BACKWARDS;
            }
            if (Keyboard::Instance().KeyPressed(KEY_LEFT) || Keyboard::Instance().KeyPressed(KEY_A))
            {
                fp_direction_flags |= FP_STRAFE_LEFT;
            }
            if (Keyboard::Instance().KeyPressed(KEY_RIGHT) || Keyboard::Instance().KeyPressed(KEY_D))
            {
                fp_direction_flags |= FP_STRAFE_RIGHT;
            }

            fp->Update(fp_direction_flags);
        }

        // Initialise the matrices
        matrix44 world_matrix = IdentityMatrix44();

        IGS->SetTransformationMatrix(OBJECT_TO_WORLD_TRANSFORM, world_matrix);

        vector3 light_direction = vector3(0.424f, -0.8f, 0.424f).normalize();

        vector3 camera, lookat, up;

        if (control_vehicle == true)
        {
            buggy->GetCameraPosition(camera, lookat, up, true);

            // Avoid camera going through the ground
            float land_height = land->GetHeight(camera.x, camera.z);
            if (land_height + 3.0f > camera.y) camera.y = land_height + 3.0f;

            cam2.setCurrentPosition(camera);
            camera = cam2.getSmooothedPosition();

            cam1.setCurrentPosition(lookat);
            lookat = cam1.getSmooothedPosition();
        }
        else
        {
            fp->GetCameraPosition(camera, lookat, up);
        }


/*
        vector3 camera(0.0f, 85.0f, 0.0f);
        vector3 lookat(255.0f, 35.0f,  255.0f);
        vector3 up(0.0f, 1.0f, 0.0f);
*/
        matrix44 view_matrix = LookAtMatrix44(camera, lookat, up);

        IGS->SetTransformationMatrix(VIEW_TRANSFORM, view_matrix);

#if defined(USE_SOUND)
        if (audio_available)
        {
            al->SetListenerPosition(camera);
            al->SetListenerOrientation(lookat, up);
            al->SetListenerVelocity(vector3(0.0f, 0.0f, 0.0f));

            for (int s = 0; s < 100; s++)
            {
                if (s != 0 && al->SourceIsPlaying(active_sources[s]) == false)
                {
                    al->DeleteSource(active_sources[s]);
                    active_sources[s] = 0;
                }
            }
        }
#endif // defined(USE_SOUND)


        matrix44 perspective_matrix = PerspectiveMatrix44(45.0f, 1.1f, 0.5f, 2000.0f);

        IGS->SetTransformationMatrix(PROJECTION_TRANSFORM, perspective_matrix);

        // Extract the 6 planes of the frustum
        ExtractFrustum(view_matrix, perspective_matrix);

        ObjectRenderer::Instance().Begin(IGS, camera, light_direction, view_matrix);


        // Set the light
        LIGHT light;
        memset(&light, 0, sizeof(LIGHT));
        light.type = DIRECTIONAL_LIGHT;

        light.direction = vector3(0.424f, -0.8f, 0.424f).normalize();

        light.ambient = COLOUR(0.5f, 0.5f, 0.5f);
        light.diffuse = COLOUR(1.0f, 1.0f, 1.0f);
        light.specular= COLOUR(1.0f, 1.0f, 1.0f);

        light.attenuation0  = 1.0f;
        light.range         = 500.0f;

        IGS->SetLight(0, light);
        IGS->EnableLight(0, true);

        IGS->SetAmbientLight(COLOUR(0.5f, 0.5f, 0.5f));

        // Ensure that this executes at least once before land->Render() is called.
        land->Update(camera);

        IGS->Clear(COLOUR(0.0f, 0.0f, 0.0f, 1.0f));

        IGS->Begin();

        IGS->FogParams(0.00065f, COLOUR(0.56f, 0.80f, 0.97f, 1.00f));
        IGS->EnableFog(false);

        IGS->SetDepthBufferMode(IGNORE_DEPTH);
        sky->Render(camera);
        IGS->SetDepthBufferMode(READ_AND_WRITE);
        IGS->EnableFog(true);
        land->Render(camera);

        IGS->EnableLighting(true);


        for (int s = 0; s < 4; s++)
        {
            matrix44 translation_sphere = ConvertToMatrix44(dBodyGetPosition(body[s]),
                dBodyGetRotation(body[s]));

            IGS->SetTransformationMatrix(OBJECT_TO_WORLD_TRANSFORM, translation_sphere);
            //DrawSphere(IGS, 3.0f, translation_sphere);
            ObjectRenderer::Instance().AddModel(thing, translation_sphere);
        }



        // Render the buggy
        int num_models = buggy->GetNumberOfModels();
        std::vector<IModel*> models;
        std::vector<matrix44> transformations;
        for (int i = 0; i < num_models; i++)
        {
            models.push_back(buggy->GetModel(i));
            transformations.push_back(buggy->GetTransformation(i));
            //ObjectRenderer::Instance().AddModel(buggy->GetModel(i), buggy->GetTransformation(i));
        }
        ObjectRenderer::Instance().AddModelGroup(models, transformations, 4);

        // Render the gun
        if (fp != NULL)
        {
            num_models = fp->GetNumberOfModels();
            for (int i = 0; i < num_models; i++)
            {
                ObjectRenderer::Instance().AddModel(fp->GetModel(i), fp->GetTransformation(i));
            }
        }

        // Update all the beasties.
        vector3 position[8];

        for (int e = 0; e < 4; e++)
        {
            position[e].x = 200 + (e/2) * 20.0f;
            position[e].z = 200 + (e%2) * 20.0f;
            position[e].y = land->GetHeight(position[e].x, position[e].z) + 2.0f;

            matrix44 trans_beastie = TranslateMatrix44(position[e].x, position[e].y, position[e].z);
            ObjectRenderer::Instance().AddModel(beastie, trans_beastie, float(GetTime()%700)/ 40.0f, beastie_pose);
        }

        ObjectRenderer::Instance().RenderObjects();

        IGS->EnableFog(false);
        sea->Render();
        IGS->EnableFog(true);

        if (g_stencil != 0)
        {
            ObjectRenderer::Instance().RenderShadows();
        }

/*
        // Deform terrain
        update_rect rect2;

        float h = float((GetTime()%20000)/100.0f);
        float* terrain_f = land->GetHeightfieldF();
        for (int x = 100; x < 140; x++)
            for (int y = 100; y < 140; y++)
                terrain_f[x+y*513] = h;

    rect2.top = 100;
    rect2.bottom = 140;
    rect2.left = 100;
    rect2.right= 140;

        land->UpdateTerrainRectangle(rect2);
*/





        //ParticleRenderer::Instance().Begin(IGS);

        long current_time = GetTime();

        int r;
        for (r = 0; r < 30; r++)
        {
            if (rockets[r].active == true)
            {
                if (current_time - rockets[r].time_fired > 10000)
                {
                    rockets[r].active = false;

                    if (rockets[r].ex != NULL)
                    {
                        delete rockets[r].ex;
                        rockets[r].ex = NULL;
                    }
                    if (rockets[r].fl != NULL)
                    {
                        delete rockets[r].fl;
                        rockets[r].fl = NULL;
                    }
                    if (rockets[r].vt != NULL)
                    {
                        delete rockets[r].vt;
                        rockets[r].vt = NULL;
                    }
                }
            }
        }


        for (r = 0; r < 30; r++)
        {
            if (rockets[r].active == true)
            {
                if (rockets[r].ex == NULL)
                {
                    vector3 current_position = rockets[r].initial_position +
                        (rockets[r].velocity * (current_time - rockets[r].time_fired)) / 1000.0f;

                    if (current_position.y < land->GetHeight(current_position.x, current_position.z))
                    {
                        vector3 ex_location = current_position + vector3(0.0, 1.0, 0.0);

                        rockets[r].ex = new Explosion(fire1_texture, fire1_texture, smoke0_texture,
                            smoke1_texture, ex_location);

                        delete rockets[r].fl;
                        rockets[r].fl = NULL;

                        rockets[r].time_exploded = GetTime();


                        // Deform terrain
                        update_rect rect;

                        sphere_subtract(land, rect, ex_location + vector3(0.0, 1.5, 0.0), 3.0f);

                        land->UpdateTerrainRectangle(rect);

                        unsigned char buffer[32 * 32 * 4];

                        for (int y = 0; y < 32; y ++)
                        {
                            for (int x = 0; x < 32; x++)
                            {
                                float d = get_distance(x, y, 0.0, 16.0, 16.0, 0.0);
                                unsigned char alpha;
                                if (d < 16) alpha = (unsigned char)(16.0 - d) * 4;
                                else alpha = 0;

                                buffer[(x + y * 32) * 4 + 0] = alpha;
                                buffer[(x + y * 32) * 4 + 1] = 0;
                                buffer[(x + y * 32) * 4 + 2] = 0;
                                buffer[(x + y * 32) * 4 + 3] = 0;
                            }
                        }

                        land->UpdateTerrainTexture(32, 32, int(ex_location.x * 4.0f) - 16, int(ex_location.z*4.0f) - 16, buffer);

                        // Sound effect
                        #if defined(USE_SOUND)
                        if (audio_available)
                        {
                            ALuint vorbsource = al->CreateSource();

                            alSourcei(vorbsource, AL_BUFFER, explosion_sample);
                            alSourcei(vorbsource, AL_LOOPING, AL_FALSE);
                            alSourcef(vorbsource, AL_ROLLOFF_FACTOR, 0.1f);

                            alSourcePlay(vorbsource);

                            al->SetSourcePosition(vorbsource, ex_location);

                            int s = 0;
                            while (active_sources[s] != 0)
                            {
                                s++;
                            };
                            active_sources[s] = vorbsource;
                        }
                        #endif // defined(USE_SOUND)

                    }
                }

                if (rockets[r].ex != NULL) rockets[r].ex->Update();
                if (rockets[r].ex != NULL) rockets[r].ex->Render();

                if (rockets[r].ex == NULL)
                {
                    if (rockets[r].vt != NULL) rockets[r].vt->Update();

                    if (rockets[r].fl != NULL) rockets[r].fl->Update();
                    if (rockets[r].fl != NULL) rockets[r].fl->Render();
                }

                // Continue rendering the remains of the vapour trail even after explosion starts
                if (rockets[r].vt != NULL) rockets[r].vt->Render();
            }
        }

        ObjectRenderer::Instance().RenderTransparentObjects();

        // Draw the HUD
        HUD_Renderer::Instance().Begin(IGS);

        DisplayText(670.0f, 0.0f, 0.4f, frame_rate_sstr.str().c_str(),
            font_texture, IGS, COLOUR(1.0f, 1.0f, 1.0f, 1.0f));

        IGS->SetTexture(0, hud_misc);

        draw_health_bar(15, /*int(float(GetTime())/200.0f)%*/11, BAR_RED);

        HUD_Renderer::Instance().Draw2D(10.0f, 3.0f, 62.0f, 19.0f,
            0.0 / 256.0, 32.0 / 256.0, 52.0 / 256.0, 48.0 / 256.0, COLOUR(1.0f, 1.0f, 0.25f, 1.0f));

/*
        static long start_fade_time = GetTime();
        long elapsed_fade_time = GetTime() - start_fade_time;

        if (elapsed_fade_time < 500)
        {
            float fade_from_black = (float)elapsed_fade_time / 500.0f;

            HUD_Renderer::Instance().Draw2D(-10.0f, -10.0f, 810.0f, 610.0f,
                65.0 / 256.0, 1.0 / 256.0, 95.0 / 256.0, 31.0 / 256.0, COLOUR(1.0f, 1.0f, 1.0f, 1.0f - fade_from_black));
        }
*/

        static bool show_get_into_vehicle_message = false;

        if (fp != NULL && get_distance(buggy->GetPosition(), fp->GetPosition()) < 10.0f)
        {
            show_get_into_vehicle_message = true;
        }

        if (show_get_into_vehicle_message)
        {
            static long start_fade_time = GetTime();
            long elapsed_fade_time = GetTime() - start_fade_time;

            if (elapsed_fade_time < 8000)
            {
                IGS->SetTexture(0, hud_misc);
                draw_message("Press [Return] to enter the vehicle", font_texture, IGS, elapsed_fade_time);
            }
        }

        static bool show_buggy_overturned_message = false;

        if (buggy->IsUpsideDown())
        {
            show_buggy_overturned_message = true;
        }

        if (show_buggy_overturned_message)
        {
            static long start_fade_time = GetTime();
            long elapsed_fade_time = GetTime() - start_fade_time;

            if (elapsed_fade_time < 8000)
            {
                IGS->SetTexture(0, hud_misc);
                draw_message("Press [R] to reset the vehicle", font_texture, IGS, elapsed_fade_time);
            }
        }


        HUD_Renderer::Instance().End();

        IGS->End();

        IGS->Display();

        SleepMS(0);
    }
    catch (std::string error_message)
    {
#if defined(WIN32)
        SDL_SysWMinfo wm_info;
        wm_info.version = SDL_version();
        if (SDL_GetWMInfo(&wm_info) == 0)
        {
            MessageBox(NULL, error_message.c_str(), "Error message",MB_OK);
        }
        else
        {
            MessageBox(wm_info.window, error_message.c_str(), "Error message",MB_OK);
        }
#endif // defined(WIN32)
        std::cout << "ERROR: " << error_message << std::endl;

        exit(1);
    }
}
