// Slickworm
// Copyright (c) 2002-2003, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "SDL_image.h"
#include "SDL.h"
#include <iostream>
#include <algorithm>
#include <sstream>
#include "graphicssystemopengl.h"
#include "surface.h"


#if defined(GL_VERSION_1_3) && !defined(WIN32)
#   define glActiveTextureARB glActiveTexture
#   define glClientActiveTextureARB glClientActiveTexture
#endif // defined(GL_VERSION_1_3)

extern int g_width, g_height;   // width and height
extern int g_stencil, g_texture;

static bool QueryExtension(char *extName)
{
    char* p = (char*) glGetString(GL_EXTENSIONS);
    char* end;
    if (p == NULL) return false;
    end = p + strlen(p);
    while (p < end)
    {
        unsigned int n = strcspn(p, " ");
        if ((strlen(extName)==n) && (strncmp(extName, p, n)==0))
        {
            return true;
        }
        p+= (n+1);
    }
    return false;
}

// Initialises the graphics system
GraphicsSystemOpenGL::GraphicsSystemOpenGL()
{
    for (int n = 0; n < 8; n++)
    {
        lights_enabled[n] = false;
    }

    // Set the viewport size
    glViewport(0,0, g_width, g_height);

    // Check if some extensions are supported.
    stencil_wrap_available = QueryExtension("GL_EXT_stencil_wrap");
    arb_texture_env_dot3_available = QueryExtension("GL_ARB_texture_env_dot3");
    arb_texture_env_combine_available = QueryExtension("GL_ARB_texture_env_combine");
/*  
    Currently disabled due to problems with some drivers.

    arb_vertex_buffer_object_available = QueryExtension("GL_ARB_vertex_buffer_object");
*/  arb_vertex_buffer_object_available = false;

#if defined(WIN32)
    // -----------------------------------------------------------------------
    // Get pointers to the address of some GL functions
    // -----------------------------------------------------------------------
    glActiveTextureARB      = (PFNGLACTIVETEXTUREARBPROC) SDL_GL_GetProcAddress ("glActiveTextureARB");
    glClientActiveTextureARB= (PFNGLCLIENTACTIVETEXTUREPROC) SDL_GL_GetProcAddress ("glClientActiveTextureARB");
    glDrawRangeElements     = (PFNGLDRAWRANGEELEMENTSPROC) SDL_GL_GetProcAddress ("glDrawRangeElements");

    glBindBufferARB         = (PFNGLBINDBUFFERARBPROC) SDL_GL_GetProcAddress("glBindBufferARB");
    glGenBuffersARB         = (PFNGLGENBUFFERSARBPROC) SDL_GL_GetProcAddress("glGenBuffersARB");
    glDeleteBuffersARB      = (PFNGLDELETEBUFFERSARBPROC) SDL_GL_GetProcAddress("glDeleteBuffersARB");
    glBufferDataARB         = (PFNGLBUFFERDATAARBPROC) SDL_GL_GetProcAddress("glBufferDataARB");
#endif

}

// Deinitialises the graphics system
GraphicsSystemOpenGL::~GraphicsSystemOpenGL()
{


}

// Clears the screen
void GraphicsSystemOpenGL::Clear(COLOUR colour)
{
    // Clear the colour buffer and depth buffer
    glClearColor( colour.r, colour.g, colour.b, colour.a);
    glClearDepth(1.0);

    if (g_stencil == 0)
    {
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    }
    else // g_stencil > 0
    {
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    }
}

// Begins the scene.
void GraphicsSystemOpenGL::Begin()
{
    // Set all the lights
    for (int n = 0; n < 8; n++)
    {
        if (lights_enabled[n] == true)
        {
            SetLight(n, lights[n]);
        }
    }
}

// Ends the scene.
void GraphicsSystemOpenGL::End()
{

}

// Displays the backbuffer onto the screen.
void GraphicsSystemOpenGL::Display()
{
    SDL_GL_SwapBuffers();
}

VERTEX_BUFFER_HANDLE GraphicsSystemOpenGL::CreateVertexList(int vertex_format, float* vertex_list,
    int number_of_vertices)
{
    int number_of_floats_per_vertex = GetNumberOfFloatsInFormat(vertex_format);

    VERTEX_BUFFER_HANDLE handle = 1;
    while(1)
    {
        if (vertex_buffers[handle].buffer == NULL)
        {
            int buffer_size = number_of_vertices * number_of_floats_per_vertex * sizeof(float);
            /*
            if (arb_vertex_buffer_object_available)
            {
                glGenBuffersARB(1, &vertex_buffers[handle].vbo_buffer);
                glBindBufferARB(GL_ARRAY_BUFFER_ARB, vertex_buffers[handle].vbo_buffer);
                glBufferDataARB(GL_ARRAY_BUFFER_ARB, buffer_size, vertex_list, GL_STATIC_DRAW_ARB);
                vertex_buffers[handle].buffer = (float*)1;
            }
            else*/ // GL_vertex_buffer_object is not supported
            {
                // Store this vertex buffer in the array of buffers.
                vertex_buffers[handle].buffer = new float[number_of_vertices * number_of_floats_per_vertex];
                memcpy(vertex_buffers[handle].buffer, vertex_list, buffer_size);
            }

            // Store some information with it
            vertex_buffers[handle].number_of_vertices = number_of_vertices;
            vertex_buffers[handle].number_of_floats_per_vertex = number_of_floats_per_vertex;
            vertex_buffers[handle].vertex_format = vertex_format;

            return handle;
        }

        handle++;

        if (handle > MAX_BUFFERS)
        {
            throw std::string("Max number of Vertex Buffers reached");
        }
    }
}

//! Deletes a vertex buffer
void GraphicsSystemOpenGL::DeleteVertexList(VERTEX_BUFFER_HANDLE handle)
{
    if (vertex_buffers[handle].buffer != NULL)
    {/*
        if (arb_vertex_buffer_object_available)
        {
            glDeleteBuffersARB(1, &vertex_buffers[handle].vbo_buffer);
        }
        else*/ // GL_vertex_buffer_object is not supported
        {
            delete[] vertex_buffers[handle].buffer;
        }
        vertex_buffers[handle].buffer = NULL;
    }
}

//! Creates an index buffer. An array of indices.
INDEX_BUFFER_HANDLE GraphicsSystemOpenGL::CreateIndexBuffer(std::vector<unsigned short> index_list)
{
    // Loop through the vector handle map looking for a free slot
    INDEX_BUFFER_HANDLE handle = 1;

    while(1)
    {
        if (index_buffers[handle].buffer == NULL)
        {/*
            if (arb_vertex_buffer_object_available)
            {
                int buffer_size = index_list.size() * sizeof(unsigned short);

                unsigned short* temp_buffer = new unsigned short[index_list.size()];
                for (unsigned int i = 0; i < index_list.size(); i++)
                {
                    temp_buffer[i] = index_list[i];
                }

                glGenBuffersARB(1, &index_buffers[handle].vbo_buffer);
                glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, index_buffers[handle].vbo_buffer);
                glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, buffer_size, temp_buffer, GL_STATIC_DRAW_ARB);
                index_buffers[handle].buffer = (unsigned short*)1;

                delete[] temp_buffer; 
            }
            else*/ // GL_vertex_buffer_object is not supported
            {
                // Store the Index buffer
                index_buffers[handle].buffer = new unsigned short[index_list.size()];
                for (unsigned int i = 0; i < index_list.size(); i++)
                {
                    index_buffers[handle].buffer[i] = index_list[i];
                }
            }

            index_buffers[handle].number_of_indices = index_list.size();

            return handle;
        }

        handle++;

        if (handle > MAX_BUFFERS)
        {
            throw std::string("Max number of Index Buffers reached");
        }
    }
}

//! Deletes an index buffer.
void GraphicsSystemOpenGL::DeleteIndexBuffer(INDEX_BUFFER_HANDLE handle)
{
    if (index_buffers[handle].buffer != NULL)
    {/*
        if (arb_vertex_buffer_object_available)
        {
            glDeleteBuffersARB(1, &vertex_buffers[handle].vbo_buffer);
        }
        else*/ // GL_vertex_buffer_object is not supported
        {
            delete[] index_buffers[handle].buffer;
        }

        index_buffers[handle].buffer = NULL;
    }
}

//! Creates a texture map from an image file
TEXTURE_HANDLE GraphicsSystemOpenGL::CreateTexture(std::string texture_filename,
    bool generate_mipmaps, bool alpha)
{
    GLuint texName;

    // Load the image from file.
    SDL_Surface *texture = IMG_Load(texture_filename.c_str());
    if (texture == NULL)
    {
        std::stringstream error_msg;
        error_msg << "Failed to load texture " << texture_filename;

        throw error_msg.str();
    }

    // Create the texture
    glGenTextures(1, &texName );
    glBindTexture(GL_TEXTURE_2D, texName);

    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

    if (generate_mipmaps)
    {
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
    }
    else
    {
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    }

    int max_width;
    glGetIntegerv(GL_MAX_TEXTURE_SIZE, &max_width);
    if (texture->w / g_texture < max_width) max_width = texture->w / g_texture;

    if (alpha == false)
    {
        if (texture->w > max_width)
        {
            unsigned char *dest = new unsigned char[max_width * max_width * 3];

            MinFilter3((unsigned char*)texture->pixels, texture->w, texture->h, dest, max_width, max_width);

            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB5, max_width, max_width, 0, GL_RGB,
                GL_UNSIGNED_BYTE, dest);

            if (generate_mipmaps)
            {
                gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB5, max_width, max_width, GL_RGB,
                    GL_UNSIGNED_BYTE, dest);
            }

            delete[] dest;
        }
        else
        {
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB5, texture->w, texture->h, 0, GL_RGB,
                GL_UNSIGNED_BYTE, texture->pixels);

            if (generate_mipmaps)
            {
                gluBuild2DMipmaps(GL_TEXTURE_2D,  GL_RGB5, texture->w, texture->h, GL_RGB,
                    GL_UNSIGNED_BYTE, texture->pixels);
            }
        }
    }
    else // (alpha == true)
    {
        if (texture->w > max_width)
        {
            unsigned char *dest = new unsigned char[max_width * max_width * 4];

            MinFilter4((unsigned char*)texture->pixels, texture->w, texture->h, dest, max_width, max_width);

            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, max_width, max_width, 0, GL_RGBA,
                GL_UNSIGNED_BYTE, dest);

            if (generate_mipmaps)
            {
                gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, max_width, max_width, GL_RGBA,
                    GL_UNSIGNED_BYTE, dest);
            }

            delete[] dest;
        }
        else
        {

            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, texture->w, texture->h, 0, GL_RGBA,
                GL_UNSIGNED_BYTE, texture->pixels);

            if (generate_mipmaps)
            {
                gluBuild2DMipmaps(GL_TEXTURE_2D,  GL_RGBA, texture->w, texture->h, GL_RGBA,
                    GL_UNSIGNED_BYTE, texture->pixels);
            }
        }
    }

    // Loop through the textures map looking for a free slot
    TEXTURE_HANDLE handle = 1;

    while(1)
    {
        if (textures[handle].used == false)
        {
            // Store this texture in the array
            textures[handle].texture_id = texName;
            textures[handle].used = true;

            return handle;
        }

        handle ++;

        if (handle > MAX_BUFFERS)
        {
            throw std::string("Max number of Textures reached");
        }
    }
}

//! Creates a texture map from a memory buffer.
TEXTURE_HANDLE GraphicsSystemOpenGL::CreateTexture(unsigned char *buffer, int width, int height, bool generate_mipmaps)
{
    GLuint texName;

    // Create the texture
    glGenTextures(1, &texName);
    glBindTexture(GL_TEXTURE_2D, texName);

    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

    if (generate_mipmaps)
    {
        glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
    }
    else
    {
        glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    }

    int max_width;
    glGetIntegerv(GL_MAX_TEXTURE_SIZE, &max_width);
    if (width / g_texture < max_width) max_width = width / g_texture;

    if (width > max_width)
    {
        unsigned char *dest = new unsigned char[max_width * max_width * 3];

        MinFilter3((unsigned char*)buffer, width, height, dest, max_width, max_width);

        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB5, max_width, max_width, 0, GL_RGB,
            GL_UNSIGNED_BYTE, dest);

        if (generate_mipmaps)
        {
            gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB5, max_width, max_width, GL_RGB,
                GL_UNSIGNED_BYTE, dest);
        }

        delete[] dest;
    }
    else
    {
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB5, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, buffer);

        if (generate_mipmaps)
        {
            gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB5, width, height, GL_RGB,
                GL_UNSIGNED_BYTE, buffer);
        }
    }

    // Loop through the textures map looking for a free slot
    TEXTURE_HANDLE handle = 1;

    while(1)
    {
        if (textures[handle].used == false)
        {
            // Store this texture in the array
            textures[handle].texture_id = texName;
            textures[handle].used = true;

            return handle;
        }

        handle ++;

        if (handle > MAX_BUFFERS)
        {
            throw std::string("Max number of Textures reached");
        }
    }
}

//! Deletes a texture
void GraphicsSystemOpenGL::DeleteTexture(TEXTURE_HANDLE handle)
{
    GLuint texName = textures[handle].texture_id;
    glDeleteTextures(1, &texName);

    textures[handle].used = false;
}

//! Sets the texture for multitexturing stage.
void GraphicsSystemOpenGL::SetTexture(int stage, TEXTURE_HANDLE handle)
{
    if (stage == 0) glActiveTextureARB(GL_TEXTURE0_ARB);
    if (stage == 1) glActiveTextureARB(GL_TEXTURE1_ARB);
    if (stage == 2) glActiveTextureARB(GL_TEXTURE2_ARB);

    if (handle != 0)
    {
        glEnable(GL_TEXTURE_2D);
        glBindTexture(GL_TEXTURE_2D, textures[handle].texture_id );
    }
    else
    {
        glDisable(GL_TEXTURE_2D);
        glDisable(GL_TEXTURE_GEN_S);
        glDisable(GL_TEXTURE_GEN_T);
    }
}

//! Sets the texture blend mode.
void GraphicsSystemOpenGL::SetMultiTextureMode(TEXTURE_BLEND_MODE mode)
{
    if (mode == NORMAL)
    {
        glActiveTextureARB(GL_TEXTURE0_ARB);
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    }
    if (mode == DETAIL_TEXTURE)
    {
        glActiveTextureARB(GL_TEXTURE0_ARB);
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

        glActiveTextureARB(GL_TEXTURE1_ARB);
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_EXT);
        glTexEnvi(GL_TEXTURE_ENV, GL_RGB_SCALE_EXT, 2);
    }
    if (mode == DOT3_BUMP)
    {
        throw std::string("DOT3 not implemented");
/*
        // UNIT 0
        // find dot product of N (stored in the texture map) and L (stored
        // as the PRIMARY_COLOR).
        glActiveTextureARB(GL_TEXTURE0_ARB);
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
        glTexEnvf(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_DOT3_RGB_ARB);

        glTexEnvf(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_TEXTURE);
        glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
        glTexEnvf(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB, GL_PRIMARY_COLOR_ARB);
        glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);

        // UNIT 1
        // modulate the base texture by N.L
        glActiveTextureARB(GL_TEXTURE1_ARB);
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
        glTexEnvf(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_MODULATE);

        glTexEnvf(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_PREVIOUS_ARB);
        glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
        glTexEnvf(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB, GL_TEXTURE);
        glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
*/
    }
}

//! Sets the texture wrapping mode
void GraphicsSystemOpenGL::SetTextureWrappingMode(int stage, TEXTURE_MODE mode)
{
    if (stage == 0) glActiveTextureARB(GL_TEXTURE0_ARB);
    if (stage == 1) glActiveTextureARB(GL_TEXTURE1_ARB);
    if (stage == 2) glActiveTextureARB(GL_TEXTURE2_ARB);

    if (mode == CLAMP_TEXTURE)
    {
        // In OpenGL these functions only apply to the currently bound texture
        glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
        glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
    }
    else // (mode == WRAP_TEXTURE)
    {
        glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
        glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );
    }
}


//! Sets the culling mode
void GraphicsSystemOpenGL::SetCullMode(cull_mode cull)
{
    switch(cull)
    {
        case CULL_CW:
            glFrontFace(GL_CCW);
            glCullFace(GL_BACK);
            glEnable(GL_CULL_FACE);
            break;
        case CULL_CCW:
            glFrontFace(GL_CW);
            glCullFace(GL_BACK);
            glEnable(GL_CULL_FACE);
            break;
        case CULL_NONE:
            glDisable(GL_CULL_FACE);
            break;
        default:
            break;
    };
}

int GraphicsSystemOpenGL::GetNumberOfFloatsInFormat(const int vertex_format)
{
    int floats_in_format = 0;

    if (vertex_format & VTX_XYZ) floats_in_format += 3;
    if (vertex_format & VTX_NORMAL) floats_in_format += 3;
    if (vertex_format & VTX_COLOUR) floats_in_format += 4;
    if (vertex_format & VTX_UV1) floats_in_format += 2;
    if (vertex_format & VTX_UV2) floats_in_format += 2;
    if (vertex_format & VTX_UV3) floats_in_format += 2;

    return floats_in_format;
}

void GraphicsSystemOpenGL::DisableAllArrays()
{
    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_NORMAL_ARRAY);
    glDisableClientState(GL_COLOR_ARRAY);
    glClientActiveTextureARB(GL_TEXTURE2_ARB);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glClientActiveTextureARB(GL_TEXTURE1_ARB);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glClientActiveTextureARB(GL_TEXTURE0_ARB);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
}

void GraphicsSystemOpenGL::EnableNeededArrays(const int vertex_format, const int number_of_floats_per_vertex,
    float* vertex_array)
{
    int floats_offset = 0;
    int floats_in_format = number_of_floats_per_vertex;

    if (vertex_format & VTX_XYZ)
    {
        glEnableClientState(GL_VERTEX_ARRAY);
        glVertexPointer(3, GL_FLOAT, floats_in_format * sizeof(GLfloat), vertex_array + floats_offset);
        floats_offset += 3;
    }
    if (vertex_format & VTX_NORMAL)
    {
        glEnableClientState(GL_NORMAL_ARRAY);
        glNormalPointer(GL_FLOAT, floats_in_format * sizeof(GLfloat), vertex_array + floats_offset);
        floats_offset += 3;
    }
    if (vertex_format & VTX_COLOUR)
    {
        glEnableClientState(GL_COLOR_ARRAY);
        glColorPointer(4, GL_FLOAT, floats_in_format * sizeof(GLfloat), vertex_array + floats_offset);
        floats_offset += 4;
    }
    if (vertex_format & VTX_UV1)
    {
        glClientActiveTextureARB(GL_TEXTURE0_ARB);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        glTexCoordPointer(2, GL_FLOAT, floats_in_format * sizeof(GLfloat), vertex_array + floats_offset);
        floats_offset += 2;
    }
    if (vertex_format & VTX_UV2)
    {
        glClientActiveTextureARB(GL_TEXTURE1_ARB);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        glTexCoordPointer(2, GL_FLOAT, floats_in_format * sizeof(GLfloat), vertex_array + floats_offset);
        floats_offset += 2;
    }
    if (vertex_format & VTX_UV3)
    {
        glClientActiveTextureARB(GL_TEXTURE2_ARB);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        glTexCoordPointer(2, GL_FLOAT, floats_in_format * sizeof(GLfloat), vertex_array + floats_offset);
        floats_offset += 2;
    }
}


void GraphicsSystemOpenGL::DrawTriangles(TRIANGLES_TYPE type, VERTEX_BUFFER_HANDLE vhandle, int vertex_count)
{
    int floats_in_format = vertex_buffers[vhandle].number_of_floats_per_vertex;
    int vertex_format = vertex_buffers[vhandle].vertex_format;
    float* vertex_array;

    /*if (arb_vertex_buffer_object_available)
    {
        glBindBufferARB(GL_ARRAY_BUFFER_ARB, vertex_buffers[vhandle].vbo_buffer);
        vertex_array = NULL;
    }
    else */// GL_vertex_buffer_object is not supported
    {
        vertex_array = vertex_buffers[vhandle].buffer;
    }

    DisableAllArrays();
    EnableNeededArrays(vertex_format, floats_in_format, vertex_array);

    if (type == TRIANGLE_LIST)
    {
        glDrawArrays(GL_TRIANGLES, 0, vertex_count);
    }
    //else if (type == TRIANGLES_
    else
    {
        throw std::string("Unsupported triangles type passed to DrawIndexedTrianglesUP()");
    }
}


void GraphicsSystemOpenGL::DrawIndexedTriangles(TRIANGLES_TYPE type, VERTEX_BUFFER_HANDLE vhandle,
    int vertex_count, INDEX_BUFFER_HANDLE ihandle, int index_count)
{
    int floats_in_format = vertex_buffers[vhandle].number_of_floats_per_vertex;
    int vertex_format = vertex_buffers[vhandle].vertex_format;
    float* vertex_array;

    /*if (arb_vertex_buffer_object_available)
    {
        glBindBufferARB(GL_ARRAY_BUFFER_ARB, vertex_buffers[vhandle].vbo_buffer);
        vertex_array = NULL;

        //glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, index_buffers[ihandle].vbo_buffer);
    }
    else */// GL_vertex_buffer_object is not supported
    {
        vertex_array = vertex_buffers[vhandle].buffer;
    }

    DisableAllArrays();
    EnableNeededArrays(vertex_format, floats_in_format, vertex_array);

    if (type == TRIANGLE_LIST)
    {/*
        if (arb_vertex_buffer_object_available)
        {
            glDrawRangeElements(GL_TRIANGLES, 0, vertex_count, index_count,
                GL_UNSIGNED_SHORT, 0);
        }
        else*/
        {
            glDrawRangeElements(GL_TRIANGLES, 0, vertex_count, index_count,
                GL_UNSIGNED_SHORT, index_buffers[ihandle].buffer);
        }
    }
    //else if (type == TRIANGLES_
    else
    {
        throw std::string("Unsupported triangles type passed to DrawIndexedTrianglesUP()");
    }
}

void GraphicsSystemOpenGL::DrawTrianglesUP(TRIANGLES_TYPE type, int vertex_format, float *vertex_array,
    int vertex_count)
{
    /*if (arb_vertex_buffer_object_available)
    {
        glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
    }*/

    int floats_in_format = GetNumberOfFloatsInFormat(vertex_format);
    DisableAllArrays();
    EnableNeededArrays(vertex_format, floats_in_format, vertex_array);

    if (type == TRIANGLE_LIST)
    {
        glDrawArrays(GL_TRIANGLES, 0, vertex_count);
    }
    //else if (type == TRIANGLES_
    else
    {
        throw std::string("Unsupported triangles type passed to DrawIndexedTrianglesUP()");
    }
}

void GraphicsSystemOpenGL::DrawIndexedTrianglesUP(TRIANGLES_TYPE type, int vertex_format, float *vertex_array,
    int vertex_count, unsigned short *index_array, int index_count)
{
    /*if (arb_vertex_buffer_object_available)
    {
        glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
    }*/

    int floats_in_format = GetNumberOfFloatsInFormat(vertex_format);
    DisableAllArrays();
    EnableNeededArrays(vertex_format, floats_in_format, vertex_array);

    if (type == TRIANGLE_LIST)
    {
        glDrawRangeElements(GL_TRIANGLES, 0, vertex_count, index_count,
            GL_UNSIGNED_SHORT, index_array);
    }
    //else if (type == TRIANGLES_
    else
    {
        throw std::string("Unsupported triangles type passed to DrawIndexedTrianglesUP()");
    }
}

matrix44 obj_to_world, view, projection;

//! Sets a transformation matrix.
void GraphicsSystemOpenGL::SetTransformationMatrix(
    TRANSFORMATION_TYPE type, const matrix44& m)
{
    GLfloat matrix[16];
    matrix[0] = m.col[0].x;
    matrix[1] = m.col[0].y;
    matrix[2] = m.col[0].z;
    matrix[3] = m.col[0].w;

    matrix[4] = m.col[1].x;
    matrix[5] = m.col[1].y;
    matrix[6] = m.col[1].z;
    matrix[7] = m.col[1].w;

    matrix[8] = m.col[2].x;
    matrix[9] = m.col[2].y;
    matrix[10]= m.col[2].z;
    matrix[11]= m.col[2].w;

    matrix[12] = m.col[3].x;
    matrix[13] = m.col[3].y;
    matrix[14] = m.col[3].z;
    matrix[15] = m.col[3].w;

    if (type == OBJECT_TO_WORLD_TRANSFORM)
    {
        obj_to_world = m;

        glMatrixMode(GL_MODELVIEW);
        glLoadMatrixf(view_matrix);
        glMultMatrixf(matrix);
    }
    else if (type == VIEW_TRANSFORM)
    {
        view = m;

        // Copy matrix to view_matrix
        memcpy(view_matrix, matrix, sizeof(GLfloat) * 16);
    }
    else if (type == PROJECTION_TRANSFORM)
    {
        projection = m;

        glMatrixMode(GL_PROJECTION);
        glLoadMatrixf(matrix);
    }
    else throw std::string("Unsupported type passed to SetTransformationMatrix");
}

//! Returns a transformation matrix.
matrix44 GraphicsSystemOpenGL::GetTransformationMatrix(TRANSFORMATION_TYPE type)
{
    if (type == OBJECT_TO_WORLD_TRANSFORM) return obj_to_world;
    else if (type == VIEW_TRANSFORM) return view;
    else if (type == PROJECTION_TRANSFORM) return projection;
    else throw std::string("Unsupported type passed to GetTransformationMatrix");
}

//! Sets which material is used for lighting.
void GraphicsSystemOpenGL::SetMaterial(MATERIAL material)
{
    GLfloat mat_ambient[] = {material.ambient.r, material.ambient.g,
        material.ambient.b, material.ambient.a };
        
    GLfloat mat_diffuse[] = {material.diffuse.r, material.diffuse.g,
        material.diffuse.b, material.diffuse.a };
    
    GLfloat mat_specular[] = { material.specular.r, material.specular.g,
        material.specular.b, material.specular.a };
        
    GLfloat mat_emissive[] = { material.emissive.r, material.emissive.g,
        material.emissive.b, material.emissive.a };

    GLfloat mat_shininess[] = { material.shiny * 128.0 };

    glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuse);
    glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
    glMaterialfv(GL_FRONT, GL_EMISSION, mat_emissive);
    glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
}

//! Sets a light
void GraphicsSystemOpenGL::SetLight(int light_number, LIGHT light)
{
    // Copy the light to the array of LIGHT structures.
    lights[light_number] = light;

    int LIGHT;

    if (light_number == 0)      LIGHT = GL_LIGHT0;
    else if (light_number == 1) LIGHT = GL_LIGHT1;
    else if (light_number == 2) LIGHT = GL_LIGHT2;
    else if (light_number == 3) LIGHT = GL_LIGHT3;
    else if (light_number == 4) LIGHT = GL_LIGHT4;
    else if (light_number == 5) LIGHT = GL_LIGHT5;
    else if (light_number == 6) LIGHT = GL_LIGHT6;
    else if (light_number == 7) LIGHT = GL_LIGHT7;
    else throw std::string("8 lights maximum has been exceeded");

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadMatrixf(view_matrix);

    GLfloat light_position[] = {    lights[light_number].position.x,
                                    lights[light_number].position.y,
                                    lights[light_number].position.z,
                                    1.0f };

    GLfloat light_direction[] = {   -lights[light_number].direction.x,
                                    -lights[light_number].direction.y,
                                    -lights[light_number].direction.z,
                                    0.0f };

    GLfloat light_ambient[] = {     lights[light_number].ambient.r,
                                    lights[light_number].ambient.g,
                                    lights[light_number].ambient.b,
                                    1.0};
    GLfloat light_diffuse[] = {     lights[light_number].diffuse.r,
                                    lights[light_number].diffuse.g,
                                    lights[light_number].diffuse.b,
                                    1.0};
    GLfloat light_specular[]= {     lights[light_number].specular.r,
                                    lights[light_number].specular.g,
                                    lights[light_number].specular.b,
                                    1.0};
    if (light.type == POINT_LIGHT)
    {
        glLightfv(LIGHT, GL_POSITION,   light_position);
    }
    else if (light.type == DIRECTIONAL_LIGHT)
    {
        glLightfv(LIGHT, GL_POSITION,   light_direction);
    }
    glLightfv(LIGHT, GL_AMBIENT,    light_ambient);
    glLightfv(LIGHT, GL_DIFFUSE,    light_diffuse);
    glLightfv(LIGHT, GL_SPECULAR,   light_specular);
    glLightf(LIGHT, GL_CONSTANT_ATTENUATION, lights[light_number].attenuation0);

    glEnable(GL_LIGHTING);
    glEnable(LIGHT);

    glPopMatrix();
}

//! Sets the colour of ambient light in the scene
void GraphicsSystemOpenGL::SetAmbientLight(COLOUR colour)
{
    GLfloat lmodel_ambient[] = { colour.r, colour.g, colour.b, colour.a };
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
}

//! Enables/Disables lighting.
void GraphicsSystemOpenGL::EnableLighting(bool enable)
{
    if (enable)
    {
        glEnable(GL_LIGHTING);
    }
    else
    {
        glDisable(GL_LIGHTING);
    }
}

//! Enables or disables a light
void GraphicsSystemOpenGL::EnableLight(int light_number, bool enable)
{
    int LIGHT;

    if (light_number == 0)      LIGHT = GL_LIGHT0;
    else if (light_number == 1) LIGHT = GL_LIGHT1;
    else if (light_number == 2) LIGHT = GL_LIGHT2;
    else if (light_number == 3) LIGHT = GL_LIGHT3;
    else if (light_number == 4) LIGHT = GL_LIGHT4;
    else if (light_number == 5) LIGHT = GL_LIGHT5;
    else if (light_number == 6) LIGHT = GL_LIGHT6;
    else if (light_number == 7) LIGHT = GL_LIGHT7;
    else throw std::string("8 lights maximum has been exceeded");
    
    if (enable == true)
    {
        glEnable(LIGHT);
        lights_enabled[light_number] = true;
    }
    else if (enable == false)
    {
        glDisable(LIGHT);
        lights_enabled[light_number] = true;
    }
}

void GraphicsSystemOpenGL::SetColour(COLOUR colour)
{
    glColor4f(colour.r, colour.g, colour.b, colour.a);
}

//! Sets the fog parameters
void GraphicsSystemOpenGL::FogParams(float density, COLOUR colour)
{
    // Set the Fog type to GL_EXP.
    glFogi(GL_FOG_MODE, GL_EXP);

    // Set the fog colour
    GLfloat fog_colour[] = {colour.r, colour.g, colour.b };

    glFogfv(GL_FOG_COLOR, fog_colour);
    glFogf(GL_FOG_DENSITY, density);
    glHint(GL_FOG_HINT, GL_DONT_CARE);
}

//! Enables or disables fog
void GraphicsSystemOpenGL::EnableFog(bool enable)
{
    if (enable == true)
    {
        glEnable(GL_FOG);
    }
    else
    {
        glDisable(GL_FOG);
    }
}

//! Sets the blend mode
void GraphicsSystemOpenGL::SetBlendMode(blend_mode blend)
{
    if (blend == SRC_ALPHA_ONE)
    {
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    }
    else if (blend == SRC_ALPHA_ONE_MINUS_SRC_ALPHA)
    {
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    }
    else if (blend == ONE_ONE)
    {
        glEnable(GL_BLEND);
        glBlendFunc(GL_ONE, GL_ONE);
    }
    else if (blend == ONE_ONE_MINUS_SRC_COLOR)
    {
        glEnable(GL_BLEND);
        glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_COLOR);
    }

    else
    {
        glDisable(GL_BLEND);
    }
}

//! Sets the depth buffer mode
void GraphicsSystemOpenGL::SetDepthBufferMode(depth_buffer_mode depth)
{
    if (depth == READ_ONLY)
    {
        glEnable(GL_DEPTH_TEST);
        glDepthMask(GL_FALSE);
        glDepthFunc(GL_LEQUAL);
    }
    else if (depth == READ_AND_WRITE)
    {
        glEnable(GL_DEPTH_TEST);
        glDepthMask(GL_TRUE);
        glDepthFunc(GL_LEQUAL);
    }
    else if (depth == IGNORE_DEPTH)
    {
        glDepthFunc(GL_ALWAYS);
        glDepthMask(GL_FALSE);
    }
}

void GraphicsSystemOpenGL::EnableWireframe(bool enable)
{
    if (enable)
    {
        glPolygonMode(GL_FRONT, GL_LINE);
        glPolygonMode(GL_BACK, GL_LINE);
    }
    else
    {
        glPolygonMode(GL_FRONT, GL_FILL);
        glPolygonMode(GL_BACK, GL_FILL);
    }
}


//! Grabs a screenshot and saves it to file.
void GraphicsSystemOpenGL::ScreenShot(std::string filename)
{
    unsigned char* pixels = new unsigned char[g_width * g_height * 3];

    glPixelStorei(GL_PACK_ALIGNMENT, 4);
    glPixelStorei(GL_PACK_ROW_LENGTH, 0);
    glPixelStorei(GL_PACK_SKIP_ROWS, 0);
    glPixelStorei(GL_PACK_SKIP_PIXELS, 0);

    glReadPixels(0,0,g_width,g_height, GL_RGB, GL_UNSIGNED_BYTE, pixels);

    FILE* out = fopen(filename.c_str(), "wb");
    fprintf(out,"P6\n%d %d\n255\n",g_width,g_height);

    for (int idx = 1; idx <= g_height; idx++)
    {
        fwrite(pixels + 3 * g_width * (g_height - idx), 3*g_width, 1, out);
    }
    fclose(out);

    delete[] pixels;
}

void GraphicsSystemOpenGL::ClearStencil()
{
    if (g_stencil != 0)
    {
        glClear(GL_STENCIL_BUFFER_BIT);
    }
}

void GraphicsSystemOpenGL::PrepareStencil()
{
    glPushAttrib(GL_ENABLE_BIT | GL_POLYGON_BIT | GL_STENCIL_BUFFER_BIT);
    glDisable(GL_LIGHTING);                                         // Turn off lighting
    glDepthMask(GL_FALSE);                                          // Turn off writing to the depth buffer
    glEnable(GL_STENCIL_TEST);                                      // Turn on stencil buffer testing
    glDisable(GL_BLEND);
    glColor4f( 1.0f, 1.0f, 1.0f, 0.0f );
    glColorMask( GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE );          // Don't draw into the colour buffer
    glStencilFunc( GL_ALWAYS, 0, ~0 );
    glStencilMask(~0);
    glDepthFunc(GL_LESS);

    SetTexture(0, 0);
    SetTexture(1, 0);
}

void GraphicsSystemOpenGL::StencilSetFirstShadowPass(shadow_mode mode)
{
    if (mode == Z_PASS_SHADOWS)
    {
        // First pass. Increase stencil value in the shadow
        SetCullMode(CULL_CW);
        glStencilOp( GL_KEEP, GL_KEEP, stencil_wrap_available ? GL_INCR_WRAP_EXT : GL_INCR );
    }
    else // (mode == Z_FAIL_SHADOWS)
    {
        // First pass.
        SetCullMode(CULL_CCW);
        glStencilOp(GL_KEEP, stencil_wrap_available ? GL_INCR_WRAP_EXT : GL_INCR, GL_KEEP);
    }
}

void GraphicsSystemOpenGL::StencilSetSecondShadowPass(shadow_mode mode)
{
    if (mode == Z_PASS_SHADOWS)
    {
        // Second pass. Decrease stencil value in the shadow
        SetCullMode(CULL_CCW);
        glStencilOp( GL_KEEP, GL_KEEP, stencil_wrap_available ? GL_DECR_WRAP_EXT : GL_DECR);
    }
    else // (mode == Z_FAIL_SHADOWS)
    {
        // Second pass.
        SetCullMode(CULL_CW);
        glStencilOp(GL_KEEP, stencil_wrap_available ? GL_DECR_WRAP_EXT : GL_DECR, GL_KEEP);
    }
}

void GraphicsSystemOpenGL::StencilShadowRectangle(COLOUR colour)
{
    glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);      // Enable rendering to colour buffer for all components

    // Draw a shadowing rectangle covering the entire screen
    glColor4f( colour.r, colour.g, colour.b, colour.a );
    glEnable(GL_BLEND);
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    glStencilFunc( GL_NOTEQUAL, 0, ~0 );
    glStencilOp( GL_KEEP, GL_KEEP, GL_KEEP );

    matrix44 old_view = GetTransformationMatrix(VIEW_TRANSFORM);
    matrix44 old_proj = GetTransformationMatrix(PROJECTION_TRANSFORM);
    matrix44 old_tran = GetTransformationMatrix(OBJECT_TO_WORLD_TRANSFORM);
    matrix44 identity = IdentityMatrix44();

    SetTransformationMatrix(VIEW_TRANSFORM, identity);
    SetTransformationMatrix(OBJECT_TO_WORLD_TRANSFORM, identity);
    matrix44 perspective_matrix = PerspectiveMatrix44(45.0f, 1.0f, 0.5f, 2000.0f);
    SetTransformationMatrix(PROJECTION_TRANSFORM, perspective_matrix);
    EnableLighting(false);
    SetDepthBufferMode(IGNORE_DEPTH);
    SetCullMode(CULL_NONE);
    glBegin( GL_TRIANGLE_STRIP );
        glVertex3f(-100.0f, 100.0f,-1.0f);
        glVertex3f(-100.0f,-100.0f,-1.0f);
        glVertex3f( 100.0f, 100.0f,-1.0f);
        glVertex3f( 100.0f,-100.0f,-1.0f);
    glEnd();
    glPopAttrib();

    glEnable(GL_DEPTH_TEST);
    glDepthMask(GL_TRUE);
    glDepthFunc(GL_LEQUAL);

    glColor4f( 1.0f, 1.0f, 1.0f, 1.0f );

    SetTransformationMatrix(VIEW_TRANSFORM, old_view);
    SetTransformationMatrix(PROJECTION_TRANSFORM, old_proj);
    SetTransformationMatrix(OBJECT_TO_WORLD_TRANSFORM, old_tran);
}



