// Slickworm
// Copyright (c) 2002-2003, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "surface.h"

// Given a source image surface of the format R8G8B8 and an already allocated
// destination surface of the format R8G8B8 of width dest_width and height
// dest_height this will scale down the source surface so that the destination
// surface is a smaller version of it (Suitable for use as a lower level mipmap
// for example).
void MinFilter3(unsigned char *source, int source_width, int source_height,
    unsigned char *dest, int dest_width, int dest_height)
{
    int stepx = source_width    / dest_width;
    int stepy = source_height   / dest_height;

    int source_pitch = source_width * 3;
    int dest_pitch = dest_width * 3;

    for (int j=0; j < dest_height; j++)
    {
        for (int i=0;i< dest_width; i++)
        {
            int redtotal  = 0;
            int greentotal= 0;
            int bluetotal = 0;

            for (int y = 0; y < stepy ;y++)
            {
                for (int x = 0; x < stepx ;x++)
                {
                    unsigned char red   = *(source + ((j * stepy) + y) * source_pitch + ((i * stepx) + x) * 3 + 0);
                    unsigned char green = *(source + ((j * stepy) + y) * source_pitch + ((i * stepx) + x) * 3 + 1);
                    unsigned char blue  = *(source + ((j * stepy) + y) * source_pitch + ((i * stepx) + x) * 3 + 2);

                    redtotal    +=red;
                    greentotal  +=green;
                    bluetotal   +=blue;
                }
            }

            int pixeltotal = (stepx * stepy);

            unsigned char red   = (redtotal  / pixeltotal);
            unsigned char green = (greentotal/ pixeltotal);
            unsigned char blue  = (bluetotal / pixeltotal);

            *(dest + j * dest_pitch + i * 3 + 0) = red;
            *(dest + j * dest_pitch + i * 3 + 1) = green;
            *(dest + j * dest_pitch + i * 3 + 2) = blue;
        }
    }
}

// Given a source image surface of the format R8G8B8A8 or A8R8G8B8 or X8R8G8B8
// and an already allocated destination surface of the format R8G8B8A8 or
// A8R8G8B8 of width dest_width and height dest_height this will scale down the
// source surface so that the destination surface is a smaller version of it
// (Suitable for use as a lower level mipmap for example).
void MinFilter4(unsigned char *source, int source_width, int source_height,
    unsigned char *dest, int dest_width, int dest_height)
{
    int stepx = source_width    / dest_width;
    int stepy = source_height   / dest_height;

    int source_pitch = source_width * 4;
    int dest_pitch = dest_width * 4;

    for (int j=0; j < dest_height; j++)
    {
        for (int i=0;i< dest_width; i++)
        {
            int redtotal  = 0;
            int greentotal= 0;
            int bluetotal = 0;
            int atotal    = 0;

            for (int y = 0; y < stepy ;y++)
            {
                for (int x = 0; x < stepx ;x++)
                {
                    unsigned char red   = *(source + ((j * stepy) + y) * source_pitch + ((i * stepx) + x) * 4 + 0);
                    unsigned char green = *(source + ((j * stepy) + y) * source_pitch + ((i * stepx) + x) * 4 + 1);
                    unsigned char blue  = *(source + ((j * stepy) + y) * source_pitch + ((i * stepx) + x) * 4 + 2);
                    unsigned char a     = *(source + ((j * stepy) + y) * source_pitch + ((i * stepx) + x) * 4 + 3);

                    redtotal    +=red;
                    greentotal  +=green;
                    bluetotal   +=blue;
                    atotal      +=a;
                }
            }

            int pixeltotal = (stepx * stepy);

            unsigned char red   = (redtotal  / pixeltotal);
            unsigned char green = (greentotal/ pixeltotal);
            unsigned char blue  = (bluetotal / pixeltotal);
            unsigned char a     = (atotal    / pixeltotal);

            *(dest + j * dest_pitch + i * 4 + 0) = red;
            *(dest + j * dest_pitch + i * 4 + 1) = green;
            *(dest + j * dest_pitch + i * 4 + 2) = blue;
            *(dest + j * dest_pitch + i * 4 + 3) = a;
        }
    }
}

// Given a source image surface of the format R8G8B8 and an already allocated
// destination surface of the format R5G6B5 and of the same width and height
// this will convert the source surface to the destination surface.
// source is an array of size width * height * 3 bytes.
// dest is an array of size width * height * 2 bytes.
void ConvertFromR8G8B8ToR5G6B5(unsigned char *source, int source_width,
    int source_height, unsigned char *dest)
{
    int source_pitch= source_width * 3;

    unsigned short *dest16 = (unsigned short *)dest;

    for (int j = 0; j < source_height; j++)
    {
        for (int i = 0; i < source_width; i++)
        {
            unsigned char red   = *(source + i * 3 + j * source_pitch + 0);
            unsigned char green = *(source + i * 3 + j * source_pitch + 1);
            unsigned char blue  = *(source + i * 3 + j * source_pitch + 2);

            unsigned short pixel = ((red >> 3) << 11) + ((green >> 2) << 5) + (blue >> 3);

            *(dest16 + i + j * source_width) = pixel;
        }
    }
}

// Given a source image surface of the format R8G8B8 and an already allocated
// destination surface of the format X1R5G5B5 and of the same width and height
// this will convert the source surface to the destination surface.
// source is an array of size width * height * 3 bytes.
// dest is an array of size width * height * 2 bytes.
void ConvertFromR8G8B8ToX1R5G5B5(unsigned char *source, int source_width,
    int source_height, unsigned char *dest)
{
    int source_pitch= source_width * 3;

    unsigned short *dest16 = (unsigned short *)dest;

    for (int j = 0; j < source_height; j++)
    {
        for (int i = 0; i < source_width; i++)
        {
            unsigned char red   = *(source + i * 3 + j * source_pitch + 0);
            unsigned char green = *(source + i * 3 + j * source_pitch + 1);
            unsigned char blue  = *(source + i * 3 + j * source_pitch + 2);

            unsigned short pixel = ((red >> 3) << 10) + ((green >> 3) << 5) + (blue >> 3);

            *(dest16 + i + j * source_width) = pixel;
        }
    }
}

